"""
Manager, environment and utils for interacting with VKontakte.

Example "configuration.json" for VKManager:

.. code-block:: json

    {
        "vk_token": "API TOKEN",
    }

Example of initializating VKManager:

.. code-block:: python

    manager1 = VKManager(
        load_configuration("vk_token", "configuration.json")
    )

"""

from .manager import VKManager, VKRequest, VKResponse
from .environment import VKEnvironment



"""Kutana - engine for developing bots."""

from kutana.logger import logger

from kutana.kutana import Kutana
from kutana.plugin import Plugin, Message, Attachment
from kutana.executor import Executor
from kutana.exceptions import ExitException

from kutana.environment import Environment

from kutana.manager.basic import BasicManager
from kutana.manager.debug import DebugManager, DebugEnvironment
from kutana.manager.vk import VKManager, VKRequest, VKResponse, VKEnvironment
from kutana.manager.tg import TGManager, TGResponse, TGEnvironment

from kutana.functions import get_path, load_configuration, import_module, \
    load_plugins, is_done

NAME = "kutana"



"""Utility functions for different purposes."""

import importlib.util
import json
import os
from os.path import dirname, isdir, join
import re

from kutana.logger import logger


def is_done(value):
    """Return True if value is None or equals to "DONE" otherwise False"""

    return value is None or value == "DONE"


def get_path(rootpath, wantedpath):
    """Return path to wantedpath relative to rootpath."""

    return join(
        dirname(rootpath),
        wantedpath
    )


def load_configuration(target, path):
    """Load specified in target key from json file in specified path."""

    with open(path, "r") as fh:
        config = json.load(fh)

    return config.get(target)


def import_module(name, path):
    """Import module from specified path with specified name.
    Return imported module.
    """

    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def load_plugin(path, plugins_list=None):
    """Load plugin from specified path. If plugins_list is specified,
    loaded plugin will be added to this lsit.
    Return loaded module or None if no plugin found.
    """

    module = import_module(path, path)

    if hasattr(module, "plugin"):
        plugins_list.append(module.plugin)

        logger.info("Loaded plugin \"%s\"", path)

        return module.plugin

    logger.warning("No plugin found in \"%s\"", path)

    return None


def load_plugins(folder):
    """Import all plugins from target folder recursively."""

    plugins_list = []

    for name in os.listdir(folder):
        path = join(folder, name)

        if isdir(path):
            plugins_list += load_plugins(path)

            continue

        if not re.match(r"^[^_].*\.py$", name):
            continue

        load_plugin(path, plugins_list)

    return plugins_list



"""Structures and classes for plugins."""

import re
import inspect
from collections import namedtuple

from kutana.functions import is_done


Message = namedtuple(
    "Message",
    "text attachments from_id peer_id date raw_update"
)

Message.__doc__ = """
Text message witch possible attachments.
"""


Attachment = namedtuple(
    "Attachment",
    "type id owner_id access_key link raw_attachment"
)

Attachment.__doc__ = """
Detailed information about attachment.
"""


Callbacks = namedtuple(
    "Callbacks",
    "normal raw"
)

Callbacks.__doc__ = """
Structure for grouping callbacks inside of :class:`.Plugin`.
"""


class Plugin():
    """Class for creating extensions for kutana engine."""

    def __init__(self, **kwargs):
        self._callbacks = Callbacks([], [])
        self._callbacks_early = Callbacks([], [])

        self._callbacks_special = []

        self._callbacks_dispose = []
        self._callbacks_startup = []

        self.priority = 400

        for k, v in kwargs.items():
            setattr(self, k, v)

    def _prepare_callbacks(self):
        """Return callbacks for registration in executor."""

        callbacks = []

        if self._callbacks_early.normal or self._callbacks_early.raw:

            async def wrapper_for_early(update, env):
                return await self._proc_update(
                    update, env, self._callbacks_early
                )

            wrapper_for_early.priority = self.priority + 200

            callbacks.append(wrapper_for_early)

        if self._callbacks.normal or self._callbacks.raw:

            async def wrapper(update, env):
                return await self._proc_update(
                    update, env, self._callbacks
                )

            wrapper.priority = self.priority

            callbacks.append(wrapper)

        callbacks.extend(self._callbacks_special)

        return callbacks

    @staticmethod
    async def _proc_update(update, env, callbacks):
        """Process update with env and target callbacks.
        If no callbacks passed raises RuntimeException.
        """

        if env.has_message():
            message = env.get_message()

        else:
            message = await env.manager.create_message(update)

            env.set_message(message)

        if message:
            callbacks_type = 0  # callbacks for messages

        else:
            callbacks_type = 1  # callbacks for raw

        inner_env = env.spawn()

        for callback in callbacks[callbacks_type]:
            if callbacks_type == 1:
                res = await callback(update, inner_env)

            else:
                res = await callback(message, inner_env)

            if res == "DONE":
                return "DONE"

    def register(self, *callbacks, early=False):
        """Register for processing updates in this plugin.

        If "early" is True, this callbacks will be executed
        before callbacks (from other plugins too) with "early=False".
        """

        if early:
            callbacks_list = self._callbacks_early.normal
        else:
            callbacks_list = self._callbacks.normal

        for callback in callbacks:
            callbacks_list.append(callback)

    def register_special(self, *callbacks, early=False):
        """Register callback for processing updates in this plugins's
        executor directly. Return decorator for registering callback.

        Arguments raw update and env is passed to callback.

        If "early" is True, this callbacks will be executed
        before callbacks (from other plugins too) with "early=False".
        """

        def _register_special(callback):
            callback.priority = self.priority + 200 * early

            self._callbacks_special.append(callback)

        for callback in callbacks:
            _register_special(callback)

        return _register_special

    def on_dispose(self):
        """Returns decorator for adding callbacks which is triggered when
        everything is going to shutdown.
        """

        def decorator(coro):
            self._callbacks_dispose.append(coro)

            return coro

        return decorator

    def on_startup(self):
        """Returns decorator for adding callbacks which is triggered
        at the startup of kutana. Decorated coroutine receives update and
        plugin environment (although this env is only accessible in plugin by
        decoraed coroutines).
        """

        def decorator(coro):
            self._callbacks_startup.append(coro)

            return coro

        return decorator

    def on_raw(self, early=False):
        """Returns decorator for adding callbacks which is triggered
        every time when update can't be turned into :class:`.Message` or
        :class:`.Attachment` object. Arguments raw update and
        :class:`.Environment` is passed to callback.

        See :func:`Plugin.register` for info about "early".
        """

        def decorator(coro):
            async def wrapper(update, env):
                if is_done(await coro(update, env)):
                    return "DONE"

            if early:
                self._callbacks_early.raw.append(wrapper)

            else:
                self._callbacks.raw.append(wrapper)

            return coro

        return decorator

    def on_text(self, *texts, early=False):
        """Returns decorator for adding callbacks which is triggered
        when the message and any of the specified text are fully matched.

        See :func:`Plugin.register` for info about "early".
        """

        if not texts:
            raise ValueError('No texts passed to "Plugin.on_text"')

        check_texts = list(text.strip().lower() for text in texts)

        def decorator(coro):
            async def wrapper(message, env):
                if message.text.strip().lower() in check_texts:
                    if is_done(await coro(message, env)):
                        return "DONE"

            self.register(wrapper, early=early)

            return coro

        return decorator

    def on_has_text(self, *texts, early=False):
        """Returns decorator for adding callbacks which is triggered
        when the message contains any of the specified texts.

        Keyword argument "found_text" can be passed to callback with text
        found in message.

        See :func:`Plugin.register` for info about "early".
        """

        check_texts = tuple(text.strip().lower() for text in texts) or ("",)

        def decorator(coro):
            signature = inspect.signature(coro)

            async def wrapper(msg, env):
                check_text = msg.text.strip().lower()

                if not check_text:
                    return "GOON"

                for text in check_texts:
                    if text not in check_text:
                        continue

                    kwargs = {}

                    if "found_text" in signature.parameters:
                        kwargs["found_text"] = text

                    if is_done(await coro(msg, env, **kwargs)):
                        return "DONE"

            self.register(wrapper, early=early)

            return coro

        return decorator

    def on_startswith_text(self, *texts, early=False):
        """Returns decorator for adding callbacks which is triggered
        when the message starts with any of the specified texts.

        Keyword arguments "body" (text after prefix), "args" (text after
        prefix splitted by spaces) and "prefix" (text before body) can
        be passed to callback.

        See :func:`Plugin.register` for info about `early`.
        """

        check_texts = tuple(text.lstrip().lower() for text in texts)

        def decorator(coro):
            signature = inspect.signature(coro)

            def search_prefix(message):
                for text in check_texts:
                    if message.startswith(text):
                        return text

                return None

            async def wrapper(msg, env):
                search_result = search_prefix(msg.text.lower())

                if search_result is None:
                    return

                kwargs = {}

                if "body" in signature.parameters:
                    kwargs["body"] = msg.text[len(search_result):].strip()

                if "args" in signature.parameters:
                    if "body" in kwargs:
                        kwargs["args"] = kwargs["body"].split()

                    else:
                        kwargs["args"] = msg.text[len(search_result):].split()

                if "prefix" in signature.parameters:
                    kwargs["prefix"] = msg.text[:len(search_result)].strip()

                if is_done(await coro(msg, env, **kwargs)):
                    return "DONE"

            self.register(wrapper, early=early)

            return coro

        return decorator

    def on_regexp_text(self, regexp, flags=0, early=False):
        """Returns decorator for adding callbacks which is triggered
        when the message matches the specified regular expression.

        Keyword argument "match" can be passed to callback with
        :class:`re.Match` object for message.

        See :func:`Plugin.register` for info about "early".
        """

        if isinstance(regexp, str):
            compiled = re.compile(regexp, flags=flags)

        else:
            compiled = regexp

        def decorator(coro):
            signature = inspect.signature(coro)

            async def wrapper(message, env):
                match = compiled.match(message.text)

                if not match:
                    return

                kwargs = {}

                if "match" in signature.parameters:
                    kwargs["match"] = match

                if is_done(await coro(message, env, **kwargs)):
                    return "DONE"

            self.register(wrapper, early=early)

            return coro

        return decorator

    def on_attachment(self, *types, early=False):
        """Returns decorator for adding callbacks which is triggered
        when the message has attachments of the specified type
        (if no types specified, then any attachments).

        See :func:`Plugin.register` for info about "early".
        """

        def decorator(coro):
            async def wrapper(message, env):
                attachments = message.attachments

                if not attachments:
                    return

                if types:
                    for a in attachments:
                        if a.type in types:
                            break
                    else:
                        return

                if is_done(await coro(message, env)):
                    return "DONE"

            self.register(wrapper, early=early)

            return coro

        return decorator



"""Class performing callbacks management."""

import sys

from kutana.logger import logger


class Executor:
    """Class performing callbacks management."""

    def __init__(self):
        self.callbacks = []
        self.error_callbacks = []

        self.startup_callbacks = []
        self.dispose_callbacks = []

        self.registered_plugins = []

    def register_plugins(self, plugins):
        """
        Register plugins' callbacks in executor.

        :param plguins: plugins for registration
        """

        for plugin in plugins:
            if hasattr(plugin, "_prepare_callbacks"):
                self.register(*plugin._prepare_callbacks())  # pylint: disable=W0212

            if hasattr(plugin, "_prepare_callbacks_error"):
                self.register(*plugin._prepare_callbacks_error(), error=True)  # pylint: disable=W0212

            self.register_startup(*plugin._callbacks_startup)  # pylint: disable=W0212
            self.register_dispose(*plugin._callbacks_dispose)  # pylint: disable=W0212

            self.registered_plugins.append(plugin)

    def register(self, *callbacks, priority=400, error=False):
        """
        Register callbacks for processing updates or errors with specified
        priority.

        :param callbacks: callbacks for registration
        :param priority: priority of callbacks
        :param error: True if registration for processing errors
        :rtype: decorator for registration
        """

        def _register(coroutine):
            callbacks = self.error_callbacks if error else self.callbacks

            callbacks.append(coroutine)

            def get_priority(callback):
                if hasattr(callback, "priority") and \
                        callback.priority is not None:
                    return -callback.priority

                return -priority

            callbacks.sort(key=get_priority)

            return coroutine

        for callback in callbacks:
            _register(callback)

        return _register

    def register_startup(self, *callbacks):
        """
        Register callbacks for startup of engine.

        :param callbacks: callbacks for registration
        :rtype: decorator for registration
        """

        def _register_startup(coroutine):
            self.startup_callbacks.append(coroutine)
            return coroutine

        for callback in callbacks:
            _register_startup(callback)

        return _register_startup

    def register_dispose(self, *callbacks):
        """
        Register callbacks for disposing resourses before shutting down
        engine.

        :param callbacks: callbacks for registration
        :rtype: decorator for registration
        """

        def _register_dispose(coroutine):
            self.dispose_callbacks.append(coroutine)
            return coroutine

        for callback in callbacks:
            _register_dispose(callback)

        return _register_dispose

    async def __call__(self, update, env):
        try:
            for callback in self.callbacks:
                if await callback(update, env) == "DONE":
                    break

        except Exception as e:  # pylint: disable=W0703
            logger.exception(
                "\"%s::%s\" on update %s from %s",
                sys.exc_info()[0].__name__, e, update, env.manager_type
            )

            env.exception = e

            if not self.error_callbacks:
                return await env.reply(
                    "Произошла ошибка! Приносим свои извинения."
                )

            for callback in self.error_callbacks:
                if await callback(update, env) == "DONE":
                    break

    async def startup(self, kutana):
        """Call initialization callbacks."""

        for callback in self.startup_callbacks:
            await callback(kutana, self.registered_plugins)

    async def dispose(self):
        """Free resources and prepare for shutdown."""

        for callback in self.dispose_callbacks:
            await callback()



"""Core class for engine."""

import asyncio

from kutana.exceptions import ExitException
from kutana.executor import Executor


class Kutana:
    """Main class for constructing engine."""

    def __init__(self, executor=None, loop=None):
        self.managers = []
        self.executor = executor or Executor()

        self.loop = loop or asyncio.get_event_loop()

        self.storage = {}

        self.running = True
        self.loops = []
        self.tasks = []

    def add_manager(self, manager):
        """Add manager to engine."""

        self.managers.append(manager)

    async def process_update(self, mngr, update):
        """Create environment and process update from manager."""

        await self.executor(update, await mngr.get_environment(update))

    def ensure_future(self, awaitable):
        """Shurtcut for asyncio.ensure_loop with active loop."""

        future = asyncio.ensure_future(awaitable, loop=self.loop)

        self.tasks.append(future)

        return future

    async def loop_for_manager(self, mngr):
        """Receive and process updates from target manager."""

        receiver = await mngr.get_receiver_coroutine_function()

        while self.running:
            for update in await receiver():
                self.ensure_future(
                    self.process_update(
                        mngr, update
                    )
                )

            await asyncio.sleep(0.01)

    def run(self):
        """Start engine."""

        asyncio.set_event_loop(self.loop)

        self.loops = []

        for manager in self.managers:
            self.loops.append(self.loop_for_manager(manager))

            awaitables = self.loop.run_until_complete(
                manager.get_background_coroutines(self.ensure_future)
            )

            for awaitable in awaitables:
                self.loops.append(awaitable)

        self.loop.run_until_complete(self.executor.startup(self))

        try:
            self.loop.run_until_complete(
                asyncio.gather(*self.loops)
            )

        except (KeyboardInterrupt, ExitException):
            pass

        finally:
            self.running = False

        self.loop.run_until_complete(self.dispose())

    async def dispose(self):
        """Free resources and prepare for shutdown."""

        await asyncio.gather(*self.tasks)

        await self.executor.dispose()

        for manager in self.managers:
            await manager.dispose()



"""Base environment for manager."""

class Environment:
    """Base environment for manager."""

    __slots__ = (
        "_has_message", "_message", "manager", "parent_environment", "peer_id",
        "exception"
    )

    def __init__(self, manager, parent_environment=None, peer_id=None):
        self.manager = manager
        self.parent_environment = parent_environment

        self.peer_id = peer_id

        self.exception = None

        self._has_message = None
        self._message = None

    @property
    def manager_type(self):
        """Return currently processed message's manager type."""

        return self.manager.type

    def spawn(self):
        """
        Create partial copy of environment with this environment as parent.
        """

        return self.__class__(
            manager=self.manager,
            parent_environment=self,
            peer_id=self.peer_id
        )

    def has_message(self):
        """
        Return True if message was created from update
        (successfully or not).
        """

        return self._has_message

    def get_message(self):
        """Return currently processed message."""

        return self._message

    def set_message(self, message):
        """
        Change message that will be passed to every sequential callbacks'
        calls. Message should be instance of :class:`.Message`
        """

        self._message = message
        self._has_message = message is not None

    async def request(self, method, **kwargs):
        """Proxy for manager's "request" method."""

        return await self.manager.request(
            method, **kwargs
        )

    async def send_message(self, message, peer_id, attachment=None, **kwargs):
        """Proxy for manager's "send_message" method."""

        return await self.manager.send_message(
            message, peer_id, attachment, **kwargs
        )

    async def reply(self, message, attachment=None, **kwargs):
        """
        Reply to currently processed message.

        :param message: message to reply with
        :param attachment: optional attachment or list of attachments
        :param kwargs: arguments for request to service
        :rtype: list with results of sending messages
        """

        return await self.manager.send_message(
            message, self.peer_id, attachment, **kwargs
        )

    async def upload_doc(self, file, **kwargs):
        """
        Upload of prepare file to be sent with :func:`.send_message`
        (or :func:`.reply`) as document. This can vary in managers'
        implementations.

        :param file: document as file or bytes
        :param kwargs: arguments for manager's method
        :rtype: :class:`.Attachment` or None
        """

        raise NotImplementedError

    async def upload_photo(self, file, **kwargs):
        """
        Upload of prepare file to be sent with :func:`.send_message`
        (or :func:`.reply`) as photo. This can vary in managers'
        implementations.

        :param file: photo as file or bytes
        :param kwargs: arguments for manager's method
        :rtype: :class:`.Attachment` or None
        """

        raise NotImplementedError

    async def get_file_from_attachment(self, attachment):
        """
        Try to download passed attachment and return it as bytes. Return None
        if error occured.

        :param attachment: :class:`.Attachment`
        :rtype: bytes or None
        """

        raise NotImplementedError
